#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""
@File    :   libs.py
@Time    :   2021/07/26 10:54:16
@Author  :   GitOPEN
@Version :   1.0
@Contact :   gitopen@gmail.com
@License :   (C)Copyright 2021-2022
@Desc    :   用于快捷导入常用的包，注意这个文件不要格式化！
"""

__version__ = "0.2.5"


# #######################################################
# ### 注意！！这个文件不能格式化，否则下面的包顺序会乱掉！
# #######################################################


# ## 精简编程时的麻烦，常用的其他包的导入
# 画图的包
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

# ## gitopenlib 本身的包的导入
from gitopenlib.helpers import mongo as gm
from gitopenlib.indicators import diversity as gd
from gitopenlib.indicators import statistics as gs
from gitopenlib.utils import basics as gb
from gitopenlib.utils import files as gf
from gitopenlib.utils import others as go
from gitopenlib.utils import plot as gp
from gitopenlib.utils import wonders as gw
from gitopenlib.utils import crawler as gc

# ## 其他常用的包
import os
import time
import sys
from collections import defaultdict
from tqdm import tqdm
import traceback

# ## 科学计算常用包
import pandas as pd
from pandas import DataFrame, Series
import numpy as np



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2020
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2020-10-29 13:38:36
# @Description :  Powered by GitOPEN

__version__ = "0.5.01"

import asyncio
import json
from pathlib import Path
import time
from typing import Callable, Iterable, List, Optional, Union

from gitopenlib.utils import basics as gb


def if_path_exist_then_backup(pathes: Union[str, List[str]]) -> bool:
    """
    检查路径是否存在，如路径存在，则备份。

    Args:
        pathes: 路径，可以是单个字符串或者字符串的列表。

    Returns:
        bool: 如果有文件被备份了，则为True，否则为False
    """
    if isinstance(pathes, str):
        pathes = [pathes]

    has_backup_files = False
    for path in pathes:
        path = Path(path)
        if path.exists():
            path.rename(
                path.with_suffix(
                    f'.{str(time.strftime("%Y%m%d_%H%M%S",time.localtime()))}{path.suffix}'
                )
            )
            has_backup_files = True
    return has_backup_files


def new_dirs(dir_paths: Union[str, List[str]]) -> List[str]:
    """
    初始化文件夹，检验文件夹的存在状态，并返回准备好的文件夹路径。

    Args:
        dir_paths: 字符串类型的全路径，可以为单个路径，也可以放入列表中，批量创建。

    Returns:
        包含创建的文件夹路径字符串的list
    """
    if isinstance(dir_paths, str):
        dir_paths = [dir_paths]

    for dir in dir_paths:
        dir_path = Path(dir)
        dir_path.mkdir(parents=True, exist_ok=True)

    return dir_paths


def get_paths_from_dir(
    dirs: Union[str, List[str]],
    types: Optional[Union[str, List[str]]] = None,
    recusive: bool = False,
) -> List[str]:
    """
    从指定目录下获取所有指定扩展名文件的路径

    Args:
        dirs: 文件夹路径（绝对路径），单个用str表示，多个用list
        types: 指定文件的扩展名，单个用str表示，多个用list，默认为None，表示所有类型文件
        recusive: 默认为False，不递归子文件夹

    Returns:
        文件路径字符串列表
    """

    if isinstance(dirs, str):
        dirs = [dirs]
    if types is not None:
        if isinstance(types, str):
            types = [types]
        rule = "**/*." if recusive else "*."
    else:
        rule = "**/*" if recusive else "*"
        types = [""]

    result = []
    for d in dirs:
        path = Path(d).resolve()
        for t in types:
            res = path.glob(rule + t)
            result.extend(list(res))

    return [str(item) for item in result]


def file_writer(
    lines: Union[str, Iterable[str]],
    dir_path: Union[str, Path],
    file_name: str = "file_writer.txt",
    mode: str = "a+",
    separator: str = "\n",
    encoding: str = "utf-8",
    backup: bool = True,
) -> None:
    """向文件中写内容

    Args:
        lines: 可以是单个字符串或者字符串的列表
        dir_path: 文件的绝对路径
        file_name: 文件名称，有默认值
        mode: 写文件的的模式，默认为 a+
        separator: 每一行末尾的分隔符，有默认值
        encoding: 文件的编码格式，默认为utf-8
        backup: 如果文件已经存在，是否对原文件进行备份，默认为True；
            如果为False，当文件已经存在的情况下，会追加到存在的文件中。
    """
    dir_path: Path = Path(dir_path) if isinstance(dir_path, str) else dir_path

    if not dir_path.exists():
        print(
            "## Warning: The dir path you submit is not exist, "
            + "it will be created automatically."
        )
        dir_path.mkdir(parents=True, exist_ok=True)

    file_path = dir_path / file_name

    if backup:
        if file_path.exists():
            file_path.rename(
                file_path.with_suffix(
                    f'.{str(time.strftime("%Y%m%d_%H%M%S",time.localtime()))}{file_path.suffix}'
                )
            )

    if isinstance(lines, str):
        lines = [lines]

    with open(file_path, mode=mode, encoding=encoding) as file:
        for line in lines:
            if separator != "":
                file.write(f"{line}{separator}")
            else:
                file.write(line)


def read_txt_by_page(
    file_path: str,
    parse_func: Callable[[list], None],
    page_size: int = 100,
    encoding: str = "utf-8",
) -> None:
    """
    从文本文件中分页读取内容

    Args:
        file_path: 文件路径
        parse_func: 每一页数据处理函数，必须定义
        page_size: 每一页数据量
        encoding: 文本文件的编码格式
    """
    with open(file=file_path, encoding=encoding, mode="r") as file:
        data = list()
        curr_page_id = 0
        start_time = time.time()
        total_time = 0.0
        print("## read_txt_by_page is starting parse the data...")
        for line in file:
            data.append(line.strip())
            if len(data) == page_size:
                parse_func(data)
                data.clear()
                end_time = time.time()
                cost_time = float(end_time - start_time)
                total_time += cost_time
                start_time = end_time
                print(f"## -{curr_page_id}- page parsed done...{[cost_time]}s")
                curr_page_id += 1

            pass
        if len(data) > 0:
            parse_func(data)
            end_time = time.time()
            cost_time = float(end_time - start_time)
            total_time += cost_time
            start_time = end_time
            print(f"## -{curr_page_id}- page parsed done...{[cost_time]}s")
            curr_page_id += 1
    print(f"## All done. Total pages: {curr_page_id}. Elapsed time: {total_time}s")


def read_txt_by_page_asyncio(
    file_path: str,
    parse_func: Callable[[list], None],
    page_size: int = 100,
    encoding: str = "utf-8",
    open_async: bool = False,
    slave_num: int = 4,
) -> None:
    """
    从文本文件中分页读取内容，具备异步io处理功能

    Args:
        file_path: 文件路径
        parse_func: 每一页数据处理函数，必须定义
        page_size: 每一页数据量
        encoding: 文本文件的编码格式
        open_async: 是否开启异步io处理数据，默认不开启
        slave_num: 执行任务的协程数目，默认为4
    """

    async def parse_(loop, chunk):
        run_func = lambda: parse_func(chunk)
        await loop.run_in_executor(None, run_func)

    with open(file=file_path, encoding=encoding, mode="r") as file:
        data = list()
        curr_page_id = 0
        start_time = time.time()
        total_time = 0.0
        print("## read_txt_by_page is starting parse the data...")
        for line in file:
            data.append(line.strip())
            if len(data) == page_size:
                if open_async:
                    loop = asyncio.get_event_loop()
                    chunks = gb.chunks(data, slave_num)
                    loop.run_until_complete(
                        asyncio.gather(*[parse_(loop, chunk) for chunk in chunks])
                    )
                    chunks.clear()
                else:
                    parse_func(data)
                # parse_func(data)
                data.clear()
                end_time = time.time()
                cost_time = float(end_time - start_time)
                total_time += cost_time
                start_time = end_time
                print(f"## -{curr_page_id}- page parsed done...{[cost_time]}s")
                curr_page_id += 1

            pass
        if len(data) > 0:
            if open_async:
                loop = asyncio.get_event_loop()
                chunks = gb.chunks(data, slave_num)
                loop.run_until_complete(
                    asyncio.gather(*[parse_(loop, chunk) for chunk in chunks])
                )
                chunks.clear()
            else:
                parse_func(data)
            # parse_func(data)
            end_time = time.time()
            cost_time = float(end_time - start_time)
            total_time += cost_time
            start_time = end_time
            print(f"## -{curr_page_id}- page parsed done...{[cost_time]}s")
            curr_page_id += 1
    print(f"## All done. Total pages: {curr_page_id}. Elapsed time: {total_time}s")


def read_content(
    file_path: Union[str, Path],
    encoding: str = "utf-8",
) -> List[str]:
    """
    从文本文件中读取内容，将内容转换为list，list的元素为每行的字符串

    Args:
        file_path : 文件路径
        encoding : 文件的编码方式

    Returns:
        每行字符串组成的list
    """
    if isinstance(file_path, str):
        file_path = Path(file_path)
    result = gb.remove_0_str(
        [line.strip() for line in file_path.read_text(encoding=encoding).split("\n")]
    )
    return result


def read_contents(
    file_pathes: List,
    encoding: str = "utf-8",
) -> List[str]:
    """
    从[多个]文本文件中读取内容，将[所有]内容转换为list，list的元素为每行的字符串

    Args:
        file_pathes : 文件路径列表
        encoding : 文件的编码方式

    Returns:
        每行字符串组成的list
    """
    result = []
    for path in file_pathes:
        result.extend(read_content(path, encoding))
    return result


def read_jsons(file_path: Union[str, Path], encoding: str = "utf-8") -> List[dict]:
    """
    从存放json的文本文件中读取内容，并转化为dict组成的list

    Args:
        file_path: 文件路径
        encoding: 文件的编码方式

    Returns:
        dict组成的list
    """
    if isinstance(file_path, str):
        file_path = Path(file_path)
    result = gb.remove_0_str(
        [line.strip() for line in file_path.read_text(encoding=encoding).split("\n")]
    )

    return [json.loads(item) for item in result]



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2021
# @Author :  GitOPEN / Big Teacher Brother
# @Email  :  gitopen@gmail.com
# @Date   :  2021-04-01 15:01:56
# @Description :  熵权法，计算指标的权重

__version__ = "0.1.0"

import math

import numpy as np
import pandas as pd
from pandas import DataFrame


def cal_weight(data: DataFrame or list, out_dict: bool = True):
    """
    使用熵权法计算指标的权重

    Args:
        data (DataFrame or list): 类型为list或者pandas的DataFrame。
        out_dict (bool): 若为True，则转化为键值对以dict类型返回；否则，以DataFrame类型返回。

    return:
        dict or DataFrame: 返回的权重值。

    """

    if isinstance(data, list):
        data = pd.DataFrame(data)

    # 去除空值
    data.dropna()

    # 归一化
    data = data.apply(
        lambda data: ((data - np.min(data)) / (np.max(data) - np.min(data)))
    )
    # 列名，也就是每个参数的名称
    columns = data.columns

    # 求k，即 -1/ln(n)
    rows = data.index.size  # 行
    cols = data.columns.size  # 列
    k = 1.0 / math.log(rows)

    # 矩阵计算
    # 信息熵
    data = np.array(data)
    lnf = [[None] * cols for i in range(rows)]
    lnf = np.array(lnf)
    for i in range(0, rows):
        for j in range(0, cols):
            if data[i][j] == 0:
                lnfij = 0.0
            else:
                p = data[i][j] / data.sum(axis=0)[j]
                lnfij = math.log(p) * p * (-k)
            lnf[i][j] = lnfij
    lnf = pd.DataFrame(lnf)
    E = lnf

    # 计算冗余度
    d = 1 - E.sum(axis=0)

    # 计算各指标的权重
    w = [[None] * 1 for i in range(cols)]
    for j in range(0, cols):
        wj = d[j] / sum(d)
        w[j] = wj

    w = pd.DataFrame(w)
    w.index = columns
    w.columns = ["weight"]
    result = w.transpose()
    if out_dict:
        for idx, row in result.iterrows():
            result = dict(row)
    return result



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2021
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2021-05-22 17:08:35
# @Description :  一些画图的相关工具函数


__version__ = "0.7.3"


from matplotlib import ticker
from matplotlib.axes import Axes
from matplotlib.offsetbox import AnchoredText
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

from gitopenlib.utils import files as gf


def save_svg(plt, path: str, dpi: int = 350, backup: bool = True):
    """保存为svg格式的矢量图。

    Parameters
    ----------
    plt : pyplot
        pyplot别名。
    path : str
        图片路径。
    dpi : int
        dpi大小，默认350。
    backup : bool
        默认为True，表示如果path存在，则备份之前的文件。
    """
    if backup:
        gf.if_path_exist_then_backup(path)
    plt.savefig(path, dpi=dpi)


def set_ax_space(plt, w: float = 0.2, h: float = 0.2):
    """设置子图间距

    Parameters
    ----------
    plt : pyplot
        pyplot别名。
    w : float
        横向的间距，值为小数，表示横向间距是子图平均宽度的百分比。
    h : float
        纵向的间距，值为小数，表示纵向间距是子图平均高度的百分比。
    """
    plt.tight_layout()
    plt.subplots_adjust(wspace=w, hspace=h)


def set_tick_integer(ax: Axes, axis: str = "both"):
    """设置横纵轴刻度值为整数

    Parameters
    ----------
    ax : Axes
        轴，可以理解为某个子图存放的位置。

    axis : str
        默认值为'both'，表示横纵轴的刻度都设置为整数；可选值维'x'或'y'。

    """
    if axis == "both":
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        ax.yaxis.set_major_locator(MaxNLocator(integer=True))
    elif axis == "x":
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))
    elif axis == "y":
        ax.yaxis.set_major_locator(MaxNLocator(integer=True))
    else:
        pass


def set_legend_outside(ax: Axes, loc: int = 3, ncol: int = 1, alpha: float = 1.0):
    """把图例放到图片的右下角；也可以调整一些参数，例如透明度，列数。"""
    ax.legend_.remove()
    legend = ax.legend(bbox_to_anchor=(1.05, 0), loc=loc, ncol=ncol, borderaxespad=0)
    legend.get_frame().set_alpha(alpha)


def set_axis_tick(ax: Axes, axis: str = "y", format="%.2f"):
    """
    横轴或者纵轴的刻度标签的格式，例如，%.2f 表示两位小数；
    %.2e 科学计数法
    """
    if axis == "x":
        ax.xaxis.set_major_formatter(ticker.FormatStrFormatter(format))
    if axis == "y":
        ax.yaxis.set_major_formatter(ticker.FormatStrFormatter(format))


def legend_text(
    ax: Axes, text: str, loc=2, fontsize: int = 10, fontcolor: str = "black"
):
    """
    在ax上添加一个仅包含文本的legend

    Args:
        ax (Axes): 子图的轴
        text (str): 要显示的文字内容
        loc (int): legend的位置，1表示右上角，2表示左上角，3表示左下角，4表示右下角
        fontsize (int): 字体大小
        fontcolor (str): 字体颜色

    Returns:
        AnchoredText: 返回AnchoredText实例
    """
    at = AnchoredText(
        text, prop=dict(size=fontsize, color=fontcolor), frameon=True, loc=loc
    )
    at.patch.set_boxstyle("round,pad=0.,rounding_size=0.2")
    ax.add_artist(at)
    return at


def set_font(fname: str = "SimHei", fsize: int = 12):
    """
    设置字体

    Args:
        fname (str): 字体的名称。
        fsize (int): 字体的大小。

    Returns:
        None
    """
    # 用来正常显示中文标签
    plt.rcParams["font.sans-serif"].insert(0, fname)
    # 用来设置字体大小
    plt.rcParams["font.size"] = fsize
    # 用来正常显示负号
    plt.rcParams["axes.unicode_minus"] = False
    plt.rcParams["mathtext.fontset"] = "cm"



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2021
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2021-02-01 10:25:51
# @Description :  一些高级功能用法

__version__ = "0.5.2"

import asyncio
from functools import wraps
from multiprocessing import Process
from time import time
from typing import Callable

from gitopenlib.utils import basics as gb


def timing(f: Callable):
    """A simple timer decorator.

    装饰器，用于统计函数的运行时间。
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        start_ = time()
        result = f(*args, **kwargs)
        end_ = time()
        print(
            f"Elapsed time # {f.__name__} # : {gb.time_formatter(end_ - start_, False)}"
        )
        return result

    return wrapper


@timing
def run_tasks_parallel(
    data: list,
    parse_func: Callable[[list], None],
    slave_num: int = 4,
):
    """
    并行处理数据的函数，加快处理速度。

    Args:
        data: list类型，需要被处理的数据。
        parse_func: 处理数据的函数，自行定义。
        slave_num: 并行任务数目。
    """

    async def parse_(loop, chunk):
        await loop.run_in_executor(None, lambda: parse_func(chunk))

    loop = asyncio.get_event_loop()
    chunks = gb.chunks(data, slave_num)
    loop.run_until_complete(asyncio.gather(*[parse_(loop, chunk) for chunk in chunks]))


@timing
def run_tasks_by_multithread(
    data: list,
    parse_func: Callable[[list], None],
    thread_num: int = 4,
):
    """
    使用多线程并行执行。

    Args:
        data: list类型，需要被处理的数据。
        parse_func: 处理数据的函数，自行定义。
        thread_num: 线程数目。
    """

    async def parse_(loop, chunk):
        await loop.run_in_executor(None, lambda: parse_func(chunk))

    loop = asyncio.get_event_loop()
    chunks = gb.chunks(data, thread_num)
    loop.run_until_complete(asyncio.gather(*[parse_(loop, chunk) for chunk in chunks]))


@timing
def run_tasks_by_multiprocess(
    data,
    parse_func: Callable[[list], None],
    process_num: int = 4,
):
    """
    使用多进程并行执行。
    **注意**：使用多进程时，parse_func需要放在py文件的顶级缩进（顶着行首），
    即，不能放到任何其他块体类。

    Args:
        data: list类型，需要被处理的数据。
        parse_func: 处理数据的函数，自行定义。
        process_num: 多进程数目。
    """
    chunks = gb.chunks(data, process_num)
    process_list = []
    for index, chunk in enumerate(chunks):
        p = Process(target=parse_func, args=(chunk,))
        p.start()
        process_list.append(p)
    for p in process_list:
        p.join()



# -*- coding:utf-8 -*-

# Copyright (c) 2022
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2022-01-15 22:47:56
# @Description :  Powered by 存放NLP常用的一些工具函数

__version__ = "0.4.3"

import random
import re
import string
from operator import itemgetter

import emoji
import jieba
import jieba.posseg as psg
from gitopenlib.libs import *
from gitopenlib.utils import nlp as gn
from jieba.analyse import TFIDF


class CutTFIDF(TFIDF):

    """对 jieba 分词的 TFIDF 类进行优化，去除单个字不能返回权重的限制。

    支持返回词汇列表对应的权重值和词性列表。
    """

    def __init__(self, idf_path=None):
        super(CutTFIDF, self).__init__(idf_path)

    def extract_tags(self, sentence, topK=100, min_length=2, allowPOS=[]):
        """切分词，并且返回词汇的 tfidf 权重和词性。

        ----------
        Parameter:
            - topK: 返回多少个top关键词。 `None` 表示所有。
            - min_length: 单词或字符的最小长度。设置为 `1` 表示单个字符。
                            如果词汇的长度小于改值，就会被过滤掉。
            - allowPOS: 允许的词性列表。例如：['ns', 'n', 'vn', 'v','nr'].
                            如果词汇的词性不在列表中，就会被过滤掉。
        """
        allowPOS = frozenset(allowPOS)
        words = self.postokenizer.cut(sentence)

        freq = {}
        flag = {}
        for w_f in words:
            w = w_f.word
            f = w_f.flag
            if len(w) < min_length:
                continue
            if allowPOS:
                if f not in allowPOS:
                    continue
            if w.lower() in self.stop_words:
                continue
            freq[w] = freq.get(w, 0.0) + 1.0
            flag[w] = f

        total = sum(freq.values())
        for k in freq:
            kw = k.word if allowPOS else k
            freq[k] *= self.idf_freq.get(kw, self.median_idf) / total

        tokens = list(freq.keys())[:topK]
        weights = list(freq.values())[:topK]
        flags = itemgetter(*tokens)(flag)
        return tokens, weights, flags


def remove_punc(text, repl=""):
    """
    默认去除替换所有半角全角符号，只留字母、数字、中文；

    或者把所有半角全角符号替换为指定的占位符(repl的值)。
    """
    rule = re.compile(r"[^a-zA-Z0-9\u4e00-\u9fa5]")
    text = rule.sub(repl, text)
    return text


def is_all_chinese(text: str):
    """
    检验是否全是中文字符（不包含标点）
    """
    text = remove_punc(text)
    for _char in text:
        if not "\u4e00" <= _char <= "\u9fa5":
            return False
    return True


def is_contains_chinese(text: str):
    """检验是否含有中文字符（不包含标点）"""
    text = remove_punc(text)
    for _char in text:
        if "\u4e00" <= _char <= "\u9fa5":
            return True
    return False


def generate_random_strs(length: int):
    """生成随机字符串"""
    return "".join(random.sample(string.digits * 5 + string.ascii_letters * 4, length))


def char_is_emoji(character):
    """判断字符是否是emoji"""
    return character in emoji.UNICODE_EMOJI


def text_has_emoji(text):
    """判断文本中是否包含emoji"""
    for character in text:
        if character in emoji.UNICODE_EMOJI:
            return True
    return False


def chinese_word_cut(
    text: str,
    custom_dict_file=None,
    stop_words: list = [],
    word_length: int = 2,
    top_k: int = 100,
    reserved_flags: list = [],
    remove_punc: bool = True,
) -> list:
    """
    对一段文本进行切分词。

    Args:
        mytext : 文本内容
        custom_dict_file : 自定义词典路径
        stop_words : 停用词列表
        word_length : 词汇的最小长度
        top_k : 保留前多少个关键词
        reserved_flags : 切分词后保留哪些词性的词汇

    Returns:
        分词列表
    """
    if custom_dict_file:
        jieba.load_userdict(custom_dict_file)
    jieba.initialize()

    token_list, weight_list, flag_list = [], [], []
    tokens, weights, flags = CutTFIDF().extract_tags(
        sentence=text, topK=top_k, min_length=word_length, allowPOS=reserved_flags
    )
    for index, token in enumerate(tokens):
        if remove_punc:
            token = gn.remove_punc(token)
            if len(token) == 0:
                continue
        if token in stop_words:
            continue
        if reserved_flags:
            if flags[index] not in reserved_flags:
                continue
        token_list.append(token)
        weight_list.append(weights[index])
        flag_list.append(flags[index])

    return token_list, weight_list, flag_list


def english_word_cut(
    text: str,
    stop_words: list = [],
    remove_punc: bool = True,
) -> list:
    """对英文句子分词（按照空格）"""
    if remove_punc:
        text = gn.remove_punc(text)
    return [word for word in text.split() if word not in set(stop_words)]


"""
if __name__ == "__main__":
    text = "喝现挤牛羊奶有感染布鲁氏菌病的风险"
    result = chinese_word_cut(
        text=text,
        custom_dict_file="/Users/sunjiajia/Works/Projects/PycharmProjects/jiaozhen/output/ML/data/custom_dict_file.txt",
        word_length=1,
        stop_words=[],
        reserved_flags=[],
    )
    print(type(result))
    print("**" * 32)
    for item in list(zip(*result)):
        print(item)

    print("**" * 32)
    res = list(zip(*result))
    res = sorted(res, key=itemgetter(1), reverse=True)
    for item in res:
        print(item)
"""



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2021
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2021-03-18 10:54:33
# @Description :  一些常用的有关 机器学习、深度学习 的通用函数

__version__ = "0.1.1"


import math
from typing import List


def softmax(data: list, decimal: int = None) -> List[float]:
    """
    归一化指数函数softmax，将一个任意实数的K维向量，压缩到另一个K维向量中，使得每一个元素
    的范围都在(0,1)之间，并且所有元素的和为1。
    """
    z_exp = [math.exp(i) for i in data]
    sum_z_exp = sum(z_exp)

    softmax = [
        round(i / sum_z_exp, decimal) if decimal is not None else (i / sum_z_exp)
        for i in z_exp
    ]
    return softmax



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2020
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2020-05-30 22:51:45
# @Description : 包含基本的文件读写，指定扩展名文件查找等基本工具


__version__ = "0.19.0"


import json
import math
import random
import time
from typing import Any, Dict, Iterable, List, Union

from gitopenlib.utils import basics as gb


def dict2object(adict: dict) -> object:
    """把dict转为类对象，实现属性的快速访问"""

    class Struct:
        def __init__(self, **entries):
            self.__dict__.update(entries)

    return Struct(**adict)


def get_keys_from_dict(adict: Dict[Any, Any]) -> List[Any]:
    """获取dict中的所有key，包括嵌套dict中的key"""
    keys = set()

    def get_keys(bdict: dict, keys: set):
        keys_ = list(bdict.keys())
        keys.update(keys_)
        values_ = list(bdict.values())
        for val in values_:
            if isinstance(val, dict):
                get_keys(val, keys)

    get_keys(adict, keys)
    keys = list(keys)
    keys.sort()
    return keys


def is_subset(a: Iterable, b: Iterable) -> bool:
    """判断a是否是b的子集，顺序不敏感"""
    return set(a).issubset(set(b))


def is_sublist(a: List, b: List) -> bool:
    """判断a是否是b的子集，顺序敏感"""
    b = iter(b)
    return all(i in b for i in a)


def list_item_getter(data: list, index: list):
    """按照index索引列表从data中拿出相应元素组成新的list"""
    return [data[idx] for idx in index]


def fmt_seconds(seconds: int or float, lang: str = "zh"):
    """把秒转换为年月日小时分钟秒。"""
    if lang == "zh":
        return time.strftime("%Y年%m月%d日 %H小时%M分钟%S秒", time.gmtime(seconds))
    else:
        return time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime(seconds))


def list_intersection(listA: list, listB: list):
    """
    求两个list的交集，listA中的哪些元素，在listB中出现了。
    注意，listA和listB的顺序不一样，交集结果不一样。
    """
    return [it for it in listA if it in listB]


def dict_extremum(data: dict, type=0) -> tuple:
    """
    找到dict中value的极值，并返回key，若极值有多个，则返回多个key

    Args:
        data: dict数据，且value必须为数值类型
        type: 1表示执行最大值操作，0表示最小值操作

    Returns:
        极值，键的列表
    """
    values = list(data.values())
    if type == 1:
        # 极大值
        ex = max(values)
    elif type == 0:
        # 极小值
        ex = min(values)
    else:
        raise Exception("The value of 'type' should only be 0 or 1.")

    # 拿到所有的key
    keys_ = [k for k, v in data.items() if v == ex]

    return ex, keys_


def list_extremum(data: list, type=0) -> tuple:
    """
    找到list中的极值，并返回索引，若极值有多个，则返回多个索引

    Args:
        data: list数据
        type: 1表示执行最大值操作，0表示最小值操作

    Returns:
        极值，索引的列表
    """
    if type == 1:
        # 极大值
        ex = max(data)
    elif type == 0:
        # 极小值
        ex = min(data)
    else:
        raise Exception("The value of 'type' should only be 0 or 1.")
    # 拿到所有索引
    idx = [id for id, item in enumerate(data) if item == ex]

    return ex, idx


def split_strip(
    strings: Union[str, List[str]],
    sep: str,
    maxsplit: int = -1,
) -> List[str]:
    """
    拆分字符串，并对列表中的元素进行 strip、去除空字符串。
    """
    if isinstance(strings, str):
        return gb.remove_0_str([it.strip() for it in strings.split(sep, maxsplit)])
    elif isinstance(strings, list):
        result = []
        for item in strings:
            item = str(item).strip()
            result.append(
                gb.remove_0_str([it.strip() for it in item.split(sep, maxsplit)])
            )
        return result


def printj(
    msg: dict,
    sort_keys: bool = False,
    beautify: bool = True,
    ensure_ascii: bool = False,
) -> None:
    """
    把dict类型的数据，格式化为json字符串输出显示。
    """
    print(
        json.dumps(msg, sort_keys=sort_keys, indent=4, ensure_ascii=ensure_ascii)
        if beautify
        else json.dumps(msg, ensure_ascii=ensure_ascii)
    )


def random_color(n: int = 1) -> Union[str, List]:
    """
    随机生成颜色代码

    Args:
        n: 生成颜色代码的数目。

    Returns:
        n为1时返回一个颜色代码字符串，n不为1时返回颜色代码列表。
    """
    if n <= 0:
        raise Exception(
            "Invalid param 'n' you provided, which sholud be equal or greater than 1."
        )
    colorArr = [
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
    ]
    colors = list()
    for i in range(n):
        color = ""
        for i in range(6):
            color += colorArr[random.randint(0, 14)]
        colors.append("#" + color)

    return colors if n > 1 else colors[0]


def filter_by_thresholds(
    data: list, thresholds: list, threshold_closed: str = None
) -> List:
    """依据阈值范围对数据进行过滤

    依据 threshold 中的规则对 data 数据进行过滤，threshold_closed
    规定区间规则的闭合状态，rl/lr 表示区间两端闭合，r表示右闭合，l表示左闭合

    Args:
        data: list类型的数据；
        thresholds: list类型的参数，元素为tuple类型；
        threshold_closed: str类型的参数，rl/lr
            表示区间两端闭合，r 表示右闭合，l 表示左闭合，None 表示两端开区间。

    Returns:
        按照区间保留下来的数据。
    """
    for th in thresholds:
        for index, value in enumerate(data):
            if threshold_closed == "rl" or threshold_closed == "lr":
                # 闭合区间，保留区间内的所有值，排除区间外的值
                if value < th[0] or value > th[1]:
                    data[index] = ""
            if threshold_closed == "r":
                # 左开右闭
                if value <= th[0] or value > th[1]:
                    data[index] = ""
            if threshold_closed == "l":
                # 左闭右开
                if value < th[0] or value >= th[1]:
                    data[index] = ""
            if threshold_closed is None:
                # 两端开区间
                if value <= th[0] or value >= th[1]:
                    data[index] = ""
        data = remove_0_str(data)

    return data


def ele2dict(data: list) -> List[dict]:
    """把元素为json字符串的list数据，转为，元素为dict类型的list数据

    Args:
        data: list类型的数据，元素为json字符串类型

    Returns:
        转换后的list数据
    """
    return [json2dict(item) for item in data]


def ele2json(data: list) -> List[str]:
    """把元素为dict类型的list数据，转为，元素为json字符串的list数据

    Args:
        data: list类型的数据，元素为dict类型

    Returns:
        转换后的list数据
    """
    return [dict2json(item) for item in data]


def dict2json(data: dict) -> str:
    """将dict转为json字符串

    Args:
        data: dict类型的数据

    Returns:
        返回json字符串
    """
    return json.dumps(data, ensure_ascii=False, separators=[",", ":"])


def json2dict(astr: str) -> dict:
    """将json字符串转为dict类型的数据对象

    Args:
        astr: json字符串转为dict类型的数据对象

    Returns:
        返回dict类型数据对象
    """
    return json.loads(astr)


def dict_sorted(data: dict, flag: int = 0, ascending: bool = True) -> dict:
    """
    对dict排序

    Args:
        data: 目标数据dict类型
        flag: 默认为0，表示按照字典的key进行排序；1表示按照value进行排序
        ascending: 默认为True，表示按照升序排序；False表示降序排序

    Returns:
        排序后的数据
    """
    return dict(sorted(data.items(), key=lambda x: x[flag], reverse=not ascending))


def list_deduplicate(data: list) -> List:
    """列表项为dict类型的去重

    Args:
        data: 目标数据，list类型

    Returns:
        去重后的数据
    """
    return [dict(t) for t in set([tuple(d.items()) for d in data])]


def sort_list(
    data: list,
    ascending: bool = True,
    flag: int = 0,
    position: int = 0,
    key: str = "",
) -> List:
    """
    对 list 进行排序

    Args:
        data: list 类型数据，item 为 dict 或者 basic type
        ascending: 默认为 True，表示升序；False 表示降序
        flag: 0 表示元素为基本类型，1 表示元素为 dict，2 表示元素为 tuple
        position: 如果元素为 basic type，position 保持默认即可；
            如果元素为 tuple，position 的数值表示以哪个 index 位置的值排序
        key: 如果元素为 dict，key 表示按照哪个 key 的 value 进行排序

    Returns:
        排序后的数据
    """
    if flag == 0:
        # basic type
        data.sort(reverse=not ascending)
    elif flag == 1:
        # dict type
        if key == "":
            key, _ = data[0].items()[0]
        data.sort(key=lambda x: x.get(key, 0), reverse=not ascending)
    elif flag == 2:
        # tuple type
        if position < 0 or position >= len(data[0]):
            position = 0
        data.sort(key=lambda x: x[position], reverse=not ascending)

    return data


def strips(string: str) -> str:
    """
    去除字符串两端的空格符和换行符，并且去除中间的换行符
    """
    return string.strip().replace("\n", "").replace("\r", "").replace("\r\n", "")


def remove_0_str(data: list) -> List:
    """
    去除list列表中为长度为0的字符串，用于字符串split后，列表中出现长度为0字符串的去除
    """
    return [item for item in data if len(str(item)) != 0]


def chunks(arr, m) -> List[list]:
    """分割列表，但是子list元素个数尽可能平均

    Args:
        arr: 待分割的list
        m: 分成几份

    Returns:
        分割后的每个子list都是返回结果list的一个元素
    """
    n = int(math.ceil(len(arr) / float(m)))
    return [arr[i : i + n] for i in range(0, len(arr), n)]


def time_formatter(seconds: int, show: bool = True):
    """将以秒为单位的时间转化为相应的分钟、小时、天。

    Args:
        seconds: 秒数
        show: 是否打印显示信息，默认为True，那么没有返回值，设置为False，不打印信息，但返回值
    Returns:
        若show为False，则有返回值，tuple中值的为天，小时，分钟，秒
    """
    d = int(seconds // 86400)
    h = int(seconds // 3600 % 24)
    m = int((seconds % 3600) // 60)
    s = round(seconds % 60, 3)

    msg = ""
    if d > 0:
        msg += f"{d} days, "
    if h > 0:
        msg += f"{h} hours, "
    if m > 0:
        msg += f"{m} minutes, "
    if s > 0:
        msg += f"{s} seconds."

    if show:
        print(msg)
    else:
        return msg



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2020
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2020-05-30 22:52:49
# @Description :  求解共现矩阵


__version__ = "0.2.1"

import csv
import math
from collections import defaultdict
from itertools import product

from sklearn.metrics.pairwise import cosine_distances, cosine_similarity

from gitopenlib.utils import basics as gb
from gitopenlib.utils import files as gf
from gitopenlib.utils import wonders


@wonders.timing
def generate_CoMatrix(
    dir_path: str,
    data: list,
    filter_num: int = 1,
    filter_tags: list = None,
    save: bool = False,
):
    """
    生成共现矩阵。

    Args:
        dir_path (str): 工作目录，生成文件的位置
        data (list): tags的列表，格式为 [["abc","bcd","cde"],["abc","cde","def"]]
        filter_num (int): 计算时只包含出现次数大于等于这个数值的tag
        filter_tags (list): 默认为None，只计算列表中包含的tag的共现矩阵；["abc", "cde"]表示只计算"abc","cde"的矩阵
        save (True): 默认为 False，不把向量保存到文件

    Returns:
        (tuple): (tags, frequency_vector, ochiia_correlation_vector, cosine_similarity_vector, cosine_distances_vector)，分别表示：标签list，共现矩阵，相关矩阵，余弦相似度矩阵，余弦距离矩阵

    """

    def save_matrix(path: str, labels: list, data: list):
        """
        保存到 csv
        """
        with open(path, "a+") as f:
            #  tsv_writer = csv.writer(f, delimiter="\t")
            tsv_writer = csv.writer(f)
            tsv_writer.writerow(["index"] + labels)
            for idx, item in enumerate(data):
                row_ = list()
                row_.append(labels[idx])
                row_.extend(item)
                tsv_writer.writerow(row_)

    # 工作目录
    BASE_DIR = gf.new_dirs(dir_path)[0]

    # kw 出现的次数统计
    kw_num_dict = defaultdict(int)
    for item in data:
        for kw in item:
            kw_num_dict[kw] += 1
    #  print(kw_num_dict)

    # 依据频次对关键词进行过滤
    kw_num_filter_dict = defaultdict(int)
    for kw, num in kw_num_dict.items():
        if num >= filter_num:
            kw_num_filter_dict[kw] = num

    kw_num_dict = kw_num_filter_dict

    # 关键词列表
    kws = list(kw_num_dict.keys())

    # 如果同时指定了 filter_num 和 filter_kws ，
    # 那么需要对两种情况下的关键词列表取交集
    if filter_tags:
        kws = list(set(kws).intersection(set(filter_tags)))
    kws = sorted(kws)

    # 两两组合出现的次数
    pair_count_dict = defaultdict(int)
    for item in data:
        pairs = list(product(item, item))
        for pair in pairs:
            if pair[0] in set(kws) and pair[1] in set(kws):
                pair_count_dict[pair] += 1

    # 若 pair 中的两个词汇一样，那么共现次数设置为0
    pair_count_dict = dict(
        [
            (pair, 0 if pair[0] == pair[1] else count)
            for pair, count in pair_count_dict.items()
        ]
    )

    # 转换为数据结构，因为每个词汇可能和多个词汇组成pair
    # 转换为以 {'kw1':{'kw2':2,'kw3':4}}的结构，
    # 表示 kw1 和 kw2 共现 2 次，和 kw3 共现4 次。
    kw_kw_dict = defaultdict(dict)
    for pair, count in pair_count_dict.items():
        kw_kw_dict[pair[0]][pair[1]] = count

    # 补全，有的kw之间共现为0，补全一下
    for kw, kw_count in kw_kw_dict.items():
        already_kws = set(kw_count.keys())
        diffs = set(kws).difference(already_kws)
        for diff in diffs:
            kw_kw_dict[kw][diff] = 0

    # 对 kw_kw_dict 排序，以及对kw的kw_dict排序
    kw_kw_dict = gb.dict_sorted(kw_kw_dict, 0)
    kw_kw_dict = dict(
        [(kw, gb.dict_sorted(kw_count, 0)) for kw, kw_count in kw_kw_dict.items()]
    )

    # 转为向量的格式
    kw_vector_dict = dict()
    for kw, kw_count in kw_kw_dict.items():
        kw_vector_dict[kw] = list(kw_count.values())

    # 共现频次 矩阵
    frequency_vector = list()

    # Ochiia系数将共词矩阵转换为相关矩阵
    # ochiia = kw1_kw2共现次数 / ( kw2频次平方根*kw2频次平方根 )
    correlation_matrix_dict = defaultdict(dict)
    for kw, kw_count in kw_kw_dict.items():
        frequency_vector.append(list(kw_count.values()))
        for kw_sub, count in kw_count.items():
            ochiia = count / (
                math.sqrt(kw_num_dict[kw]) * math.sqrt(kw_num_dict[kw_sub])
            )
            correlation_matrix_dict[kw][kw_sub] = ochiia

    # 相关矩阵
    ochiia_correlation_vector = [
        list(val.values()) for kw, val in correlation_matrix_dict.items()
    ]

    # 计算余弦相似度
    cosine_similarity_vector = cosine_similarity(frequency_vector)

    # 计算余弦距离
    cosine_distances_vector = cosine_distances(frequency_vector)

    # 保存为 csv 矩阵格式
    if save:
        for vector, name in zip(
            [
                frequency_vector,
                ochiia_correlation_vector,
                cosine_similarity_vector,
                cosine_distances_vector,
            ],
            [
                "frequency_vector",
                "ochiia_correlation_vector",
                "cosine_similarity_vector",
                "cosine_distances_vector",
            ],
        ):
            save_matrix(
                "{}/{}.csv".format(BASE_DIR, name).replace("_vector", "_matrix"),
                kws,
                vector,
            )

    print("...done...")

    return (
        kws,
        frequency_vector,
        ochiia_correlation_vector,
        cosine_similarity_vector,
        cosine_distances_vector,
    )


#  if __name__ == "__main__":
#      data = [["abc", "bcd", "cde"], ["abc", "cde", "def"]]
#      generate_CoMatrix(dir_path="./", data=data, filter_num=1, filter_tags=None)
#      pass



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2021
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2021-10-26 14:48:44
# @Description :  存放一些其他工具函数


__version__ = "0.1.5"

import sys
import time
import traceback
from email.mime.text import MIMEText
from smtplib import SMTP_SSL

from gitopenlib.utils import basics as gb


def sendTaskOK(
    to_addr="sunjiajiacn@qq.com",
    username="gitopen@sina.cn",
    host="smtp.sina.cn",
    host_port=465,
    password=None,
    subject=None,
    message=None,
    func_main=None,
):
    """
    发送任务执行完毕通知邮件。

    Args:
        to_addr: 邮件接收地址
        username: 发件人地址
        host: smtp地址（SSL）
        host_port: smtp端口（SSL）
        password: 帐户密码
        subject: 邮件主题描述
        message: 邮件内容
        func_main: 回调方法，启用这个功能，可以将任务开始和结束时间都统计
    """
    if func_main is None:
        now_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())

        # 邮件内容
        if message is None or len(message) == 0:
            message = "<p>任务执行完毕....</p>" + "<p>{}</p>".format(now_time)

        msg = MIMEText(message, "html", _charset="utf-8")

        # 邮件主题描述
        if subject is None or len(subject) == 0:
            subject = "【任务执行完毕】({})".format(now_time)
        msg["Subject"] = subject
        msg["From"] = username
        msg["To"] = to_addr

    else:
        is_success = True
        start_time = time.time()
        start_ftime = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        # 执行回调主函数，也可能报错，但是报错的话，也算任务执行完毕，因此也要发送邮件，
        # 不过是通知任务失败的邮件
        try:
            func_main()
        except Exception as e:
            is_success = False
            traceback.print_exc()
            error_info = traceback.format_exc()
        spend_time = time.time() - start_time
        spend_time = round(spend_time, 3)
        end_ftime = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())

        # 邮件内容
        if message is None or len(message) == 0:
            message = (
                "<p><h2>任务成功！</h2></p>"
                + "<p>任务开始时间：</p>"
                + "<p>{}</p>".format(start_ftime)
                + "<p>任务结束时间：</p>"
                + "<p>{}</p>".format(end_ftime)
                + "<p>任务耗时：</p>"
                + "<p>{}</p>".format(gb.fmt_seconds(spend_time))
            )
            if not is_success:
                message = message.replace(
                    "<p><h2>任务成功！</h2></p>",
                    "<p><h2>任务失败！</h2></p><p><b>错误信息：{}</b></p>".format(error_info),
                )

        msg = MIMEText(message, "html", _charset="utf-8")

        # 邮件主题描述
        if subject is None or len(subject) == 0:
            #  subject = "【任务执行完毕】({})".format(now_time)
            subject = "【任务成功】开始：{}，结束：{}，耗时：{}".format(
                start_ftime, end_ftime, gb.fmt_seconds(spend_time)
            )
            if not is_success:
                subject = subject.replace("成功", "失败")
        msg["Subject"] = subject
        msg["From"] = username
        msg["To"] = to_addr

    # 发送邮件
    with SMTP_SSL(host=host, port=host_port) as smtp:
        # 登录发送邮件服务器
        smtp.login(user=username, password=password)
        # 实际发送、接收邮件配置
        smtp.sendmail(from_addr=username, to_addrs=to_addr, msg=msg.as_string())



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2021
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2021-03-16 16:07:34
# @Description :  使用k-mean++算法进行聚类

__version__ = "0.1.0"

import json
import os

import joblib
import matplotlib.pyplot as plt
import numpy as np
from gitopenlib.utils import files as gf
from gitopenlib.utils import wonders
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from sklearn.cluster import KMeans
from sklearn.metrics import calinski_harabasz_score, silhouette_score


@wonders.timing
def run_kmeans(K_range: list, data: list, tags: list, base_dir: str):
    """
    data 是一个 list 类型，其中的每一个 sublist 都和 labels 中的标签对应

    Args:
        max_K (list): K值试验范围
        data (list): 特征数据集，元素是list类型的特征数据
        tags (list): 每一个特征数据对应的标签
        base_dir (str): 工作目录，在该目录下生成一系列的文件和聚类结果

    """
    BASE_DIR = gf.new_dirs(base_dir)[0]

    Ks = list(range(K_range[0], K_range[1] + 1))
    models = list()
    sil_coeff_list = list()
    bss_list = list()
    chs_list = list()
    sse_list = list()

    for k_ in Ks:
        # # 创建存放结果的文件夹
        print(f"#1# the n_clusters is: {k_}")
        base_k_dir = gf.new_dirs(BASE_DIR + f"clusters/{k_}/")[0]

        # # 进行聚类
        random_state = 10
        kmeans = KMeans(n_clusters=k_, random_state=random_state).fit(X=data)
        models.append(kmeans)

        # # 评价指标，经实验观察，当以下评价指标的线型图中，
        # # 首次出现曲率最大的转折点，就是最优的聚类类别数目
        # 使用Silhouette Coefficient 评价，轮廓系数
        # Silhouette Coefficient是对象与其自身簇（内聚力）相比与其他簇（分离）相似程度的度量。
        # 值从-1到+1，其中高值表示对象与其自己的簇很好地匹配并且与相邻簇不匹配。
        # 如果大多数对象具有高值，则聚类结果是合适的。 如果许多点具有低值或负值，则聚类效果不好，可能具有太多或太少的簇。
        # 最大值的那个位置就是聚类最好的位置
        sil_coeff = silhouette_score(data, kmeans.labels_, metric="euclidean")
        sil_coeff_list.append(sil_coeff)

        # 使用between_ss，组内平方误差和法
        # 是组间平方和与总的距离平方和的商
        # 越小越好
        bss = sum_of_square_scores(
            data, kmeans.labels_, kmeans.cluster_centers_, kmeans.n_clusters
        )
        bss_list.append(bss)

        # 使用 calinski_harabaz_score 方法评价聚类效果的好坏,大概是类间距除以类内距，因此这个值越大越好
        # 越大越好
        chs = calinski_harabasz_score(data, kmeans.labels_)
        chs_list.append(chs)

        # 使用 SSE 均方误差
        # SSE越接近于0，说明模型选择和拟合更好，数据预测也越成功
        sse = kmeans.inertia_
        sse_list.append(sse)

        # 将聚类结果的一些参数写入文件中，方便后续检验和观察
        msg = (
            "#" * 23
            + "\n"
            + "n_clusters: {}\n".format(k_)
            + "cluster_centers_: {}\n".format(kmeans.cluster_centers_)
            + "sil_coeff: "
            + str(sil_coeff)
            + "\n"
            + "between_ss / total_ss: "
            + str(bss)
            + "\n"
            + "chs: "
            + str(chs)
            + "\n"
            + "sse: "
            + str(sse)
            + "\n"
        )
        gf.file_writer(msg, base_k_dir, "run_cluster_result.txt")

    # 分析
    # 评价指标的线型图
    fig: Figure = plt.figure()
    ax: Axes = fig.add_subplot(221)
    ax.plot(
        Ks, sil_coeff_list, marker="o", markersize=4, label="Silhouette Coefficient"
    )
    ax.legend()

    ax: Axes = fig.add_subplot(222)
    ax.plot(Ks, bss_list, marker="o", markersize=4, label="Between SS")
    ax.legend()

    ax: Axes = fig.add_subplot(223)
    ax.plot(Ks, chs_list, marker="o", markersize=4, label="Calinski Harabaz Score")
    ax.legend()

    ax: Axes = fig.add_subplot(224)
    ax.plot(Ks, sse_list, marker="o", markersize=4, label="SSE")

    ax.legend()

    plt.show()

    # 让用户判断哪个K值是合适的
    user_command = input("请输入您选择的K值，若输入N，则直接退出:\n")
    if user_command == "N" or user_command == "n":
        os._exit(0)

    # 捕捉异常
    try:
        k_best = int(user_command)
    except ValueError as e:
        print(e)
        os._exit(1)

    # 得到模型
    km = models[Ks.index(k_best)]
    # 分类 类别
    labels_ = km.labels_
    # 质心，质心分类号 与 质心标签的关系数据
    centers = km.cluster_centers_
    # 求出与每个质心（中心）距离最近的一个 样本点
    center_sample_points = list()
    # 求质心最近样本点所属的聚类类别
    center_sample_labels = list()
    # 求质心最近样本点的tag
    center_sample_tags = list()
    for center in centers:
        dist_list = list()
        for idx, val in enumerate(data):
            dist = np.linalg.norm(center - np.array(val))
            dist_list.append(dist)
        dist_min = min(dist_list)
        dist_min_idx = dist_list.index(dist_min)
        point = data[dist_min_idx]
        center_sample_points.append(list(point))
        label = labels_[dist_min_idx]
        center_sample_labels.append(label)
        tag = tags[dist_min_idx]
        center_sample_tags.append(tag)

    # 如果是二维的数据，那么画出来聚类结果
    # 可视化结果
    # 画出所有样例点，属于同一类的绘制同样的颜色
    if len(data[0]) == 2:
        colors = plt.cm.Spectral(np.linspace(0, 1, k_best))
        for idx, val in enumerate(data):
            plt.plot(
                val[0],
                val[1],
                marker="o",
                color=colors[int(labels_[idx])],
                markeredgecolor="black",
                markersize=6,
            )

        # 画出质点，用特殊图型
        for i in range(k_best):
            plt.plot(
                centers[i][0],
                centers[i][1],
                marker="p",
                markeredgecolor="red",
                markersize=10,
            )

        # 画出质心最近点，用特殊图型，五角星
        for i in range(k_best):
            plt.plot(
                center_sample_points[i][0],
                center_sample_points[i][1],
                marker="*",
                markeredgecolor="red",
                markersize=10,
            )
        plt.show()

    # 将聚类好的数据进行格式化整理
    # 保存模型
    joblib.dump(km, BASE_DIR + "/km_model.pkl")
    # 载入模型
    #  joblib.load(BASE_DIR + "/km_model.pkl")

    # 保存质心、质心最近点，质心最近点类别
    # {"centers":[], "center_sample":[],"label":[]}
    center_dict = {
        "centers": centers.tolist(),
        "center_sample": center_sample_points,
        "label": [int(label) for label in center_sample_labels],
    }
    gf.file_writer(
        json.dumps(center_dict, ensure_ascii=False), BASE_DIR, "center_data.txt"
    )

    # 保存质心、质心最近点，质心最近点类别
    # {"centers":[], "center_sample":[],"label":[]}
    center_sample_tags_dict = {
        "center_sample_tags": center_sample_tags,
        "label": [int(label) for label in center_sample_labels],
    }
    gf.file_writer(
        json.dumps(center_sample_tags_dict, ensure_ascii=False),
        BASE_DIR,
        "center_sample_tags_data.txt",
    )

    # 对于一个聚类类别，将数据格式化为如下格式
    # {"center_tag":xxx,"label":xxxx,"tags":[..., ..., ...]}
    cluster_dict = {"clusters": list()}
    for csp, csl in zip(center_sample_points, center_sample_labels):
        cluster = dict()

        tag_idx = int(data.index(csp))
        center_tag = tags[tag_idx]

        cluster["center_tag"] = dict()
        cluster["center_tag"]["tag"] = center_tag
        cluster["center_tag"]["data"] = csp
        cluster["label"] = int(csl)
        cluster["tags"] = list()
        for index_, label_ in enumerate(labels_):
            if label_ == csl:
                tag_ = tags[index_]
                data_ = list(data[index_])
                cluster["tags"].append({"tag": tag_, "data": data_})

        cluster_dict["clusters"].append(cluster)

    gf.file_writer(
        json.dumps(cluster_dict, ensure_ascii=False), BASE_DIR, "cluster_data.txt"
    )

    # 保存 labels ，方便使用
    labels = [int(it) for it in labels_.tolist()]
    gf.file_writer(labels, BASE_DIR, "labels.txt")

    print("#done ...")

    return labels, center_dict, cluster_dict


def sum_of_square_scores(original_data, predict_labels, cluster_centers, n_clusters):
    """
    定义between_SS / total_SS 的计算方法
    """
    avg = np.mean(original_data, axis=0)
    dist = np.power(original_data - avg, 2)
    total_ss = np.sum(dist)
    within_squares = np.zeros((n_clusters, len(original_data[0])))
    for i in range(0, len(original_data)):
        cluster = predict_labels[i]
        within_squares[cluster] += np.power(
            original_data[i] - cluster_centers[cluster], 2
        )
    within_ss = np.sum(within_squares)
    return (total_ss - within_ss) / total_ss


#  if __name__ == "__main__":
#      import random
#
#      n = 300
#
#      labels = [f"A{i}" for i in range(n)]
#      data = [[random.randint(1, n), random.randint(1, n)] for i in range(n)]
#      print(data)
#      print(labels)
#      run_kmeans([2, 20], data, labels, "/Users/sunjiajia/Downloads/test/cluster/")



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2020
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2020-11-05 10:23:57
# @Description :  一系列统计学相关的计算函数


__version__ = "0.19.7"

import math
from collections import Counter
from typing import List, Optional, Union

import numpy as np
from gitopenlib.utils import basics as gb
from scipy import stats


def divide_bins(data: list, bins: int = 30, fmt: str = "count"):
    """
    把一个数据列表，分成多少个区间，并统计落在每个区间中的数值数目或百分比。

    主要用于手动画直方图，例如：

    ```
    color = ["#FFFFFF"] * len(bins)
    plt.bar(
        [it[0] for it in bins],
        percents,
        width=bins[1][1] - bins[1][0],
        align="edge",
        color=color,
        alpha=1,
        edgecolor="#4D79C8",
    )
    ```


    Args:
        data: 数据列表
        bins: 区间格式
        fmt: count表示统计区间数值的数目，percent表示统计区间数值数目的百分比

    Returns:
        返回值有两个，第一个是区间列表，第二个是统计结果列表
    """
    # 划分为区间
    bin_edges = divide_interval(data, bins)

    # 在判断值属于某个区间的时候，采取与python区间相同的方法：前闭后开
    res_dict = dict()

    for it in data:
        for idx, bin in enumerate(bin_edges):
            if bin[0] not in res_dict:
                res_dict[bin[0]] = 0
            if idx == len(bin_edges) - 1:
                if it >= bin[0] and it <= bin[1]:
                    res_dict[bin[0]] += 1
            else:
                if it >= bin[0] and it < bin[1]:
                    res_dict[bin[0]] += 1

    res_dict = gb.dict_sorted(res_dict)
    if fmt == "count":
        counts = list(res_dict.values())
        return bin_edges, counts
    elif fmt == "percent":
        data_len = len(data)
        percents = [round(it / data_len, 5) for it in res_dict.values()]
        return bin_edges, percents
    else:
        raise Exception("fmt's value should be 'count' or 'percent'.")


def divide_interval(
    data: List[Union[int, float]], number: int, decimal: int = None
) -> List[list]:
    """把列表中数据，划分为区间

    Args:
        data: 列表类型数据，元素为整型或者浮点型。
        number: 划分的区间数目，整数类型。
        decimal: 区间端点的小数位数，默认不保留小数。

    Returns:
        返回值为列表类型，元素为划分的区间，tuple类型。

    """
    if number == 0:
        raise Exception("the number of interval can't be 0.")
    max_, min_ = max(data), min(data)
    poor = max_ - min_
    width = poor / number

    result = []
    interval = list()

    for i in range(number + 1):
        right = min_ + width * i
        if decimal:
            right = round(right, decimal)
        interval.append(right)
        if len(interval) == 2:
            result.append(tuple(interval))
            temp = tuple(interval)[-1]
            interval.clear()
            interval.append(temp)

    return result


def calculate_hist_bins2(data: list, k: float = 1.5) -> float:
    """求解画柱状图时的 bins 的数值，原理是 Freedman–Diaconis rule。"""
    iqr = calculate_IQR(data, k)[-1]
    bin_with = 2 * iqr * (1 / math.pow(len(data), 1 / 3))
    bins = (max(data) - min(data)) / bin_with
    return bins


def calculate_CDF(data: Union[list, dict]) -> dict:
    """计算 Cumulative Distribution Function (CDF)

    Args:
        data: 数据。

    Returns:
        每个key对应计算的累计频率。

    """

    if isinstance(data, list):
        data = dict(Counter(data))

    data = gb.dict_sorted(data)

    keys = list(data.keys())
    values = list(data.values())
    sum_ = sum(values)

    new_values = list()
    for index, item in enumerate(data):
        new_values.append(sum(values[: index + 1]) / sum_)

    return dict(zip(keys, new_values))


def calculate_CCDF(data: Union[list, dict]) -> dict:
    """计算 Complementary Cumulative Distribution Function (CCDF)

    Args:
        data: 数据。

    Returns:
        每个key对应计算的互补累计频率。

    """
    return {k: 1 - v for k, v in calculate_CDF(data).items()}


def calculate_PMF(data: Union[list, dict]) -> dict:
    """计算 Probability Mass Function (PMF)

    Args:
        data: 数据。

    Returns:
        每个key对应计算的单个频率。

    """

    if isinstance(data, list):
        data = dict(Counter(data))

    data = gb.dict_sorted(data)

    keys = list(data.keys())
    values = list(data.values())
    sum_ = sum(values)

    new_values = [it / sum_ for it in values]

    return dict(zip(keys, new_values))


def calculate_percent(data: dict, mode: str = "gte") -> None:
    """计算Counter形式的百分比。

    当 mode 为 gte 时，计算 大于等于 key 的占比；
    当 mode 为 lte 时，计算 小于等于 key 的占比。
    将结果打印到控制台。

    Args:
        data: Counter形式的数据；
        mode: 可选值为 gte 和 lte。

    Returns:
        None
    """
    data = gb.dict_sorted(data)
    x = list(data.keys())
    y = list(data.values())
    if isinstance(y[0], list):
        y = [sum(v) for v in y]
    total_count = sum(y)

    if mode == "gte":
        for flag in x:
            count = 0
            for i, k in enumerate(x):
                if k >= flag:
                    count += y[i]
            msg = f"key->{flag}->count->{count}->gte percent->{round(count/total_count,5)*100}%"
            print(msg)
    elif mode == "lte":
        res = calculate_CDF(data)
        for key, value in res.items():
            msg = f"key->{key}->count->{data[key]}->lte percent->{round(value,5)*100}%"
            print(msg)
    else:
        raise Exception("This operation has not been implemented.")


def curve_fit(x: np.array, y: np.array, deg: int) -> tuple:
    """
    对x,y进行曲线拟合。

    使用方法：
    1. 先画出散点图，观察数据的大致形状（如3次函数）；
    2. 使用该函数进行拟合，deg设置为3；
    3. 得到拟合后的x对应的y_fit值、R^2、已经方程。

    Args:
        x (np.array): 自变量x的值，可用list数据转为np.array。
        y (np.array): 因变量y的值，可用list数据转为np.array。

    Returns:
        y_fit (np.array): 拟合曲线对应的y值。
        r2 (np.float64): 拟合效果评价指标。
        aa (str): 方程的表现形式。
    """
    parameter = np.polyfit(x, y, deg)
    p = np.poly1d(parameter)
    print(p)
    aa = ""
    for i in range(deg + 1):

        bb = parameter[i]

        sup = deg - i
        if i != 0:
            if bb < 0:
                aa = list(aa)
                aa[-1] = "-"
                aa = "".join(aa)
                bb = 0 - bb
        if sup != 0:
            aa += "{:.3g}".format(bb) + " $x^{" + str(sup) + "}$ " + "+"
        else:
            aa += "{:.3g}".format(bb)

    r2 = round(np.corrcoef(y, p(x))[0, 1] ** 2, 3)
    y_fit = p(x)
    aa = "$y_{fit}$ = " + aa
    return y_fit, r2, aa


def normal_distribution_values(u: Union[int, float], std: Union[int, float]) -> tuple:
    """生成均值为u标准差为std的正态分布数据

    用处：当样本数据集确定（list），可以求出均值、标准差，
    生成相应的符合正态分布的y值，画出正态分布曲线，
    用于比较样本集与正态分布的区别，帮助判断


    Args:
        u: 平均值。
        std: 标准差。

    Returns:
        x (list): x轴的数据
        y (list): y轴的数据

    """
    x = np.linspace(u - 3 * std, u + 3 * std, 100)
    y = np.exp(-((x - u) ** 2) / (2 * std ** 2)) / (math.sqrt(2 * math.pi) * std)
    return x, y


def Cdf(t, x) -> float:
    """累计分布函数，就是值到其在分布中百分等级的映射。

    计算给定x的CDF(x)，就是计算样本中小于等于x的值的比例。
    在画图时，样本的CDF是一个阶跃函数（图形形状类似向上的阶梯）。

    Args:
        t (list): 样本集
        x (int or float): 给定值

    Returns:
        (float): 概率，小于等于x的值的概率

    """
    count = 0.0
    for value in t:
        if value <= x:
            count += 1.0
    prob = count / len(t)
    return prob


def percentile_rank(scores: list, your_score: Union[int, float]) -> Union[int, float]:
    """获取百分等级

    百分等级就是原始分数不 高于你的人在全部考试人数中所占的比例再乘以100。
    如果你 在 90 百分位数，那就是说你比 90% 的人成绩好，或者至少不比 90% 的考试人员差。

    Args:
        scores (list): 分数列表
        your_score (int or float): 考察对象的分数

    Returns:
        (int or float): 百分位数
    """
    count = 0
    for score in scores:
        if score <= your_score:
            count += 1
    percentile_rank = 100.0 * count / len(scores)
    return percentile_rank


def percentile(scores, prank) -> Union[int, float]:
    """根据百分等级，找到百分位数

    Args:
        scores (list): 分数列表
        prank (list): 百分等级

    Returns:
        (int or float): 百分位数

    """
    scores.sort()
    for score in scores:
        if percentile_rank(scores, score) >= prank:
            return score


def percentile2(scores, prank) -> Union[int, float]:
    """根据百分等级，找到百分位数，优化版

    Args:
        scores (list): 分数列表
        prank (list): 百分等级

    Returns:
        (int or float): 百分位数

    """
    scores.sort()
    index = percentile_rank * (len(scores) - 1) / 100
    return scores[index]


def get_high_low_threshold(minimum_count: int) -> float:
    """
    获取关键词的高低频临界值。高频低频词界分公式是 Donohue 根据齐普夫第二定律。
    依据的论文：
    Donohue J C. Understanding Scientific Literatures—A Bibliometric Approach [ M] . Cambridge:The MIT Press, 1973 :49 –50 .

    Args:
        minimum_count (int): 频次为1的关键词的数量。

    Returns:
        float : 临界值是浮点型。
    """
    return (-1 + math.sqrt(1 + 8 * minimum_count)) / 2


def get_extremum(data: Optional[dict or list], type: str = "max"):
    """获取最大值和最小值"""
    if isinstance(data, dict):
        keys, values = list(data.keys()), list(data.values())
    elif isinstance(data, list):
        keys, values = [key for key in range(len(data))], data
    else:
        raise Exception("please check the data parameter and its type.")

    if type == "max":
        extremum = max(values)
    elif type == "min":
        extremum = min(values)
    else:
        raise Exception("please check the type parameter.")

    res_keys, res_values = list(), list()
    for k, v in zip(keys, values):
        if type == "max":
            if v >= extremum:
                res_keys.append(k)
                res_values.append(v)
        elif type == "min":
            if v <= extremum:
                res_keys.append(k)
                res_values.append(v)
    return res_keys, res_values


def coefficient_of_variation(data: list) -> float:
    """计算变异系数，又称离散系数，是一个衡量数据离散程度的、没有量纲的统计量。

    通常用来比较两组量纲差异明显的数据的离散程度，例如粉丝数量差距显著的社交媒体帐号的推文点赞量的离散程度。

    """
    mean = np.mean(data)  # 计算平均值
    std = np.std(data, ddof=0)  # 计算标准差
    cv = std / mean
    return cv


def kurtosis_skewness(data: list):
    """求一组数据的峰度 kurtosis 和 偏度 Skewness。

    正态分布的峰度为3，若峰度大于0，则说明该组数据的分布曲线相较于正态分布更加陡峭；若峰度小于0，则说明该组数据的分布曲线相较于正态分布更加平缓。换言之，峰度越大，则数据在靠近均值的部分分布得越多，在距离均值较远的部分分布得较少。
    峰度可以用来衡量风险，同样的方差下，峰度越高，越容易取极端值，风险越大。
    峰度可以帮助衡量众数的统计学意义，峰度越高，众数在描述该组数据的集中趋势的作用就越高。

    正态分布的偏度为0，用来衡量数据分布的对称性，
    当偏度小于0时，称为 Negative Skew，出现左侧长尾，
    当偏度大于0时，称为 Positive Skew，出现右侧长尾，
    当偏度的绝对值过大时，长尾一侧出现极端值的可能性较高。

    """
    arr_ = np.asarray(data)

    k_ = stats.kurtosis(arr_)
    s_ = stats.skew(arr_)

    return k_, s_


def KL_divergence(p: list, q: list):
    """计算KL散度，两者越相似，KL散度越小。

    KL散度满足非负性
    KL散度是不对称的，交换P、Q的位置将得到不同结果。

    """
    p_arr = np.asarray(p)
    q_arr = np.asarray(q)

    return stats.entropy(p_arr, q_arr)


def JS_divergence(p: list, q: list):
    """计算JS散度，两者越相似，JS散度越小。

    JS散度的取值范围在0-1之间，完全相同时为0
    JS散度是对称的

    """
    p_arr = np.asarray(p)
    q_arr = np.asarray(q)

    M = (p_arr + q_arr) / 2

    return 0.5 * stats.entropy(p, M) + 0.5 * stats.entropy(q, M)


def calculate_hist_bins1(N: int, mode: int = 0) -> tuple:
    """计算柱状图的bins，依据 Sturges' rule 和 Rice's rule 两种方法。

    参见 http://onlinestatbook.com/2/graphing_distributions/histograms.html

    Args:
        N (int): 数据的数目
        mode (int): Sturges' rule 的计算方法选择，0 是 1+log2(N)，1 是 1+3.3log10(N)

    Returns:
        两种方法求 hist bins 的值

    """

    s_rule = (1 + np.log2(N)) if mode == 0 else (1 + 3.3 * np.log10(N))
    r_rule = 2 * np.power(N, 1 / 3)

    return round(s_rule), round(r_rule)


def counter2percenter(data: dict) -> dict:
    """convert counter dict to percenter dict.

    将 key/value 分别为数值、数值频次的字典类型数据，
    转换为，key/value 分别为数值、数值频次百分比的字典类型数据。

    Args:
        data (dict): dict类型的数据，key/value分别为数值、数值频次。

    Returns:
        dict: dict类型的数据，key/value分别为数值、数值频次百分比。
    """
    sum_ = sum(list(data.values()))

    return dict([(key, value / sum_) for key, value in data.items()])


def calculate_IQR(data: list, k: float = 1.5) -> tuple:
    """计算四分位距（IQR, Interquartile range）

    Args:
        data (list): list类型的数据，每个元素为 int 或 float 类型。
        k (float): 表示异常系数，k = 1.5表示中度异常，k = 3表示极度异常。

    Returns:
        tuple: 顺序为：upper whisker, upper quartile, median, lower quartile, lower whisker, IQR
    """
    Percentile = np.percentile(data, [0, 25, 50, 75, 100])
    upper_quartile = Percentile[3]
    lower_quartile = Percentile[1]
    IQR = upper_quartile - lower_quartile
    upper_whisker = upper_quartile + IQR * k
    lower_whisker = lower_quartile - IQR * k
    median = Percentile[2]

    return upper_whisker, upper_quartile, median, lower_quartile, lower_whisker, IQR


def filter_outliers_by_IQR(data: list, k: float = 1.5) -> tuple:
    """
    使用IQR对异常值进行检测，又称为 Tukey's test。

    Args:
        data (list): list类型的数据，每个元素为 int 或 float 类型。
        k (float): 表示异常系数，k = 1.5表示中度异常，k = 3表示极度异常。

    Returns:
        tuple: upper outliers, lower outliers

    """
    (
        upper_whisker,
        upper_quartile,
        median,
        lower_quartile,
        lower_whisker,
        IQR,
    ) = calculate_IQR(data, k=k)
    print(" upper_whisker : ", upper_whisker)
    print(" lower_whisker : ", lower_whisker)
    upper_outliers = [item for item in data if item > upper_whisker]
    lower_outliers = [item for item in data if item < lower_whisker]

    return upper_outliers, lower_outliers


def remove_outliers_by_IQR(data: list, k: float = 1.5) -> list:
    """
    使用IQR从数据中移除异常值，返回移除异常值的数据

    Args:
        data (list): list类型的数据，每个元素为 int 或 float 类型。
        k (float): 表示异常系数，k = 1.5表示中度异常，k = 3表示极度异常。

    Returns:
        list: 移除异常值后的数据
    """
    (
        upper_whisker,
        upper_quartile,
        median,
        lower_quartile,
        lower_whisker,
        IQR,
    ) = calculate_IQR(data, k=k)

    new_data = [
        item for item in data if item <= upper_whisker and item >= lower_whisker
    ]
    return new_data


def calculate_list_count_percent(data: list, decimals: Optional[int] = None) -> tuple:
    """计算list中元素的次数和百分比

    Args:
        data (list): list类型数据。
        decimals (None or int): 保留的小数位数，默认为None，则不进行位数操作；如果>=0，则保留相应的小数位数。

    Returns:
        tuple: 元组第一个元素为次数统计结果，第二个元素为百分比统计结果。

    """
    counter = dict(Counter(data))

    total_count = sum(list(counter.values()))

    percenter = dict(
        [
            (
                key,
                value / total_count
                if decimals is None
                else round(value / total_count, decimals),
            )
            for key, value in counter.items()
        ]
    )
    return counter, percenter


def smooth_ma(data: Optional[list or np.array], window_size=3):
    """
    平滑后的数组的长度为：len(data) - window_size + 1

    Args:
        data (list or np.array): 原始数组。
        window_size (int): 窗口大小。

    Returns:
        np.array: 平滑后的数组
    """
    cumsum_vec = np.cumsum(np.insert(data, 0, 0))
    ma_vec = (cumsum_vec[window_size:] - cumsum_vec[:-window_size]) / window_size
    return ma_vec


def smooth_matlab(data: Optional[list or np.array], window_size: int = 3):
    """
    这个方法是matlab中平滑函数的python实现。它的窗口只能为奇数。
    平滑后的数组的长度为：len(data)

    Args:
        data (list or np.array): 原始数组。
        window_size (int): 窗口大小。

    Returns:
        np.array: 平滑后的数组。

    """

    # a: NumPy 1-D array containing the data to be smoothed
    # WSZ: smoothing window size needs, which must be odd number,
    # as in the original MATLAB implementation
    out0 = np.convolve(data, np.ones(window_size, dtype=int), "valid") / window_size
    r = np.arange(1, window_size - 1, 2)
    start = np.cumsum(data[: window_size - 1])[::2] / r
    stop = (np.cumsum(data[:-window_size:-1])[::2] / r)[::-1]
    return np.concatenate((start, out0, stop))


def normalization(data: Optional[list or np.array], decimals: None or int = None):
    """
    对一系列数据进行归一化处理，对原始数据进行线性变换把数据映射到[0,1]之间。

    Args:
        data (list or np.array): 需要被归一化的数据。
        decimals (None or int): 归一化后，数值保留的精度（小数位数），默认为None，不开启

    Returns:
        np.array: 处理后的数据
    """
    _range = np.max(data) - np.min(data)
    ret = (data - np.min(data)) / _range
    return ret if decimals is None else np.around(ret, decimals=decimals)


def standardization(data: Optional[list or np.array], decimals: None or int = None):
    """
    对一系列数据进行标准化处理，常用的方法是z-score标准化，处理后数据均值为0，标准差为1。

    Args:
        data (list or np.array): 需要被标准化的数据。
        decimals (None or int): 归一化后，数值保留的精度（小数位数），默认为None，不开启

    Returns:
        np.array: 处理后的数据
    """
    mu = np.mean(data, axis=0)
    sigma = np.std(data, axis=0)
    ret = (data - mu) / sigma
    return ret if decimals is None else np.around(ret, decimals=decimals)


def zero_centered(data: Optional[list or np.array], decimals: None or int = None):
    """
    对一系列数据进行零均值化，即zero-centered。对数据的平移的一个过程，之后数据的中心点为(0,0)。

    Args:
        data (list or np.array): 需要被零均值化的数据。
        decimals (None or int): 数值保留的精度（小数位数），默认为None，不开启

    Returns:
        np.array: 处理后的数据
    """
    mu = np.mean(data, axis=0)
    ret = data - mu
    return ret if decimals is None else np.around(ret, decimals=decimals)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2021
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2021-12-06 14:10:44
# @Description :  有关爬虫的一些工具函数


__version__ = "0.0.2"


import random
from pathlib import Path

from fake_useragent import UserAgent
from gitopenlib.utils import files as gf


def update_useragent():
    uas = set()
    for i in range(1000):
        ua = UserAgent()
        uas.add(ua.random)

    gf.file_writer(uas, "./ua/", "ua.txt", backup=False)


def get_ua():
    """随机获取一个 User-Agent"""
    ua_path = "./ua/ua.txt"
    if not Path(ua_path).exists():
        update_useragent()
    uas = gf.read_content(ua_path)
    idx = random.randint(0, len(uas) - 1)
    return uas[idx]



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2021
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2021-01-11 18:55:11
# @Description :  存放常用的工具类

__version__ = "0.1.0"

from functools import partial


class F(partial):
    """实现管道。

    examples:
        >>> range(10) | F(filter, lambda x: x % 2) | F(sum)
        >>> your_dict | F(lambda d: {k: v for k, v in d.items() if v})
    """

    def __ror__(self, other):
        if isinstance(other, tuple):
            return self(*other)
        return self(other)



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2021
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2021-08-23 21:13:01
# @Description :  网络请求相关的封装，还未测试，仅仅写上去


import asyncio

import aiohttp


async def get(url: str, params: dict = None, **kwargs):
    async with aiohttp.ClientSession().get(url, params=params, **kwargs) as response:
        assert response.status == 200


async def post(url: str, data: bytes, **kwargs):
    async with aiohttp.ClientSession().post(url, data=data, **kwargs) as response:
        assert response.status == 200
        return response



#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Copyright (c) 2021
# @Author :  GitOPEN
# @Email  :  gitopen@gmail.com
# @Date   :  2021-03-25 14:35:21
# @Description :  Powered by GitOPEN


__version__ = "0.1.0"


import json
from pathlib import Path
from typing import Callable

from gitopenlib.utils import basics as gb
from gitopenlib.utils import files as gf


def parse_refworks(dir_path: str, parse_func: Callable[[list], None] = None):
    """
    对RefWorks的题录数据进行格式化，保存到json文件中，或者使用parse_func直接处理。

    Args:
        dir_path (str): 存放 txt 文件类型的RefWorks数据目录；
        parse_func (Callable): 处理函数；

    Returns:
        None
    """
    file_paths = gf.get_paths_from_dir(dirs=dir_path, types="txt")
    for path in file_paths:
        path = Path(path)
        file_name = path.name
        save_dir = str(path.parent) + "/"

        data = list()
        item = dict()

        f = open(path, "r")
        for row in f:
            row = str(row).strip()
            if row == "":
                continue

            subs = row.split(" ", 1)
            flag = subs[0].strip()
            if len(flag) == 2:
                if flag in ["A1", "AD", "K1"]:
                    subs[1] = gb.remove_0_str(subs[1].split(";"))
                item[subs[0]] = subs[1]
                if flag == "DS":
                    data.append(json.dumps(item, ensure_ascii=False))
                    item.clear()
        f.close()

        # save data to json format file.
        if parse_func is None:
            save_dir = gf.new_dirs(save_dir + "json/")[0]
            gf.file_writer(lines=data, dir_path=save_dir, file_name=f"{file_name}.json")
        # parse data by custom function.
        else:
            parse_func(data)
